import json
import logging
from typing import Dict, List

from aiohttp import ClientError, web
from aiohttp_security import check_authorized
from aiohttp_security.api import AUTZ_KEY

from .client import Permission


logger = logging.getLogger(__name__)


async def check_permissions(
    request: web.Request, permissions: List[Permission]
) -> None:
    user_name = await check_authorized(request)
    auth_policy = request.config_dict.get(AUTZ_KEY)
    if not auth_policy:
        raise RuntimeError("Auth policy not configured")

    try:
        missing = await auth_policy.get_missing_permissions(user_name, permissions)
    except ClientError as e:
        # re-wrap in order not to expose the client
        raise RuntimeError(e) from e

    if missing:
        payload = {"missing": [_permission_to_primitive(p) for p in missing]}
        raise web.HTTPForbidden(
            text=json.dumps(payload), content_type="application/json"
        )


def _permission_to_primitive(perm: Permission) -> Dict[str, str]:
    return {"uri": perm.uri, "action": perm.action}



from dataclasses import dataclass


@dataclass(frozen=True)
class BearerAuth:
    """Result of parsing the Bearer HTTP Authorization header.

    Analogous to `aiohttp.helpers.BasicAuth`.
    """

    token: str

    @classmethod
    def decode(cls, header_value: str) -> "BearerAuth":
        try:
            auth_scheme, token = header_value.split(" ", 1)
        except ValueError:
            raise ValueError("No credentials")
        if auth_scheme.lower() != "bearer":
            raise ValueError("Unexpected authorization scheme")
        if not token:
            raise ValueError("No credentials")
        return cls(token=token)

    def encode(self) -> str:
        return "Bearer " + self.token



import pytest

from neuro_auth_client.bearer_auth import BearerAuth


class TestBearerAuth:
    def test_decode_unexpected_scheme(self) -> None:
        with pytest.raises(ValueError, match="Unexpected authorization scheme"):
            BearerAuth.decode("Basic credentials")

    @pytest.mark.parametrize("header_value", ("Bearer", "Bearer "))
    def test_decode_no_credentials(self, header_value: str) -> None:
        with pytest.raises(ValueError, match="No credentials"):
            BearerAuth.decode(header_value)

    @pytest.mark.parametrize("token", ("token", "to ken"))
    def test_decode(self, token: str) -> None:
        auth = BearerAuth.decode("Bearer " + token)
        assert auth == BearerAuth(token=token)

    def test_encode(self) -> None:
        assert BearerAuth(token="token").encode() == "Bearer token"



from enum import Enum
from typing import Any, Optional, Sequence

from aiohttp.client_exceptions import ClientResponseError
from aiohttp.hdrs import AUTHORIZATION
from aiohttp.helpers import BasicAuth
from aiohttp.web import Application, Request, Response
from aiohttp_security import AbstractAuthorizationPolicy, AbstractIdentityPolicy, setup
from jose import jwt
from jose.exceptions import JWTError

from .bearer_auth import BearerAuth
from .client import AuthClient, Permission, User


JWT_IDENTITY_CLAIM = "https://platform.neuromation.io/user"
JWT_IDENTITY_CLAIM_OPTIONS = ("identity", JWT_IDENTITY_CLAIM)


class AuthScheme(str, Enum):
    BASIC = "basic"
    BEARER = "bearer"


class IdentityPolicy(AbstractIdentityPolicy):
    def __init__(self, auth_scheme: AuthScheme = AuthScheme.BEARER) -> None:
        self._auth_scheme = auth_scheme

    async def identify(self, request: Request) -> Optional[str]:
        auth_header_value = request.headers.get(AUTHORIZATION)

        if auth_header_value is None:
            return None

        if self._auth_scheme == AuthScheme.BASIC:
            identity = BasicAuth.decode(auth_header_value).password
        else:
            identity = BearerAuth.decode(auth_header_value).token

        return identity

    async def remember(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        pass

    async def forget(
        self, request: Request, response: Response
    ) -> None:  # pragma: no cover
        pass


class AuthPolicy(AbstractAuthorizationPolicy):
    def __init__(self, auth_client: AuthClient) -> None:
        self._auth_client = auth_client

    def get_user_name_from_identity(self, identity: str) -> Optional[str]:
        try:
            claims = jwt.get_unverified_claims(identity)
            for identity_claim in JWT_IDENTITY_CLAIM_OPTIONS:
                if identity_claim in claims:
                    return claims[identity_claim]
            return None
        except JWTError:
            return None

    async def authorized_user(self, identity: str) -> Optional[User]:
        """Retrieve authorized user object (works same as authorized_userid)

        Return the user object identified by the identity
        or 'None' if no user exists related to the identity.
        """
        name = self.get_user_name_from_identity(identity)
        if not name:
            return None
        try:
            # NOTE: here we make a call to the auth service on behalf of the
            # actual user, not a service.
            return await self._auth_client.get_user(name, token=identity)
        except ClientResponseError:
            return None

    async def authorized_userid(self, identity: str) -> Optional[str]:
        user = await self.authorized_user(identity)
        if user:
            return user.name
        return None

    async def permits(
        self, identity: str, _: str, context: Sequence[Permission]
    ) -> bool:
        name = self.get_user_name_from_identity(identity)
        if not name:
            return False
        return await self._auth_client.check_user_permissions(name, context)

    async def get_missing_permissions(
        self, user_name: str, permissions: Sequence[Permission]
    ) -> Sequence[Permission]:
        return await self._auth_client.get_missing_permissions(user_name, permissions)


async def setup_security(
    app: Application,
    auth_client: AuthClient,
    auth_scheme: AuthScheme = AuthScheme.BEARER,
) -> None:  # pragma: no cover
    identity_policy = IdentityPolicy(auth_scheme=auth_scheme)
    auth_policy = AuthPolicy(auth_client=auth_client)
    setup(app, identity_policy, auth_policy)
